import threading



class FramerError(Exception):
    def __init__(self, *args):
        super().__init__(*args)
class FrameError(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class Framer():
    def __init__(self):
        self._code = []
        self._vars = {}
        self._aliases = {}
        self._lock = threading.RLock()
    
    def var(self, name:str, value):
        self._vars[name] = value
        return self

    def op(self, a, b, operator: str = '+'):
        result_var = f'__tmp_{len(self._code)}'
        res = f'{a} {operator} {b}'
        code_line = f'{result_var} = {res}'
        self._code.append(code_line)
        return result_var
    
    def execute(self):
        final_code = "\n".join(self._code)
        local_scope = self._vars.copy()
        compiled = compile(final_code, '<string>', 'exec')
        exec(compiled, {}, local_scope)
        return local_scope.get(f"__tmp_{len(self._code)-1}") 
    
    def get_thread_safe(self, name):
        with self._lock:
            return self._vars.get(self._aliases[name])
    
    def _new_code_line(self, line: str): self._code.append(line)

    def __enter__(self):
        System.last_framer = System.framer
        System.framer = self
        return self
    def __exit__(self, *args, **kwargs): pass


class System:
    framer: Framer = Framer()
    last_framer: Framer = framer
    framers: dict = {'basic': Framer(), 'temp': Framer()}
    def match(condition: str, true_block: str, false_block: str = None, framer = None):
        framer = System.framer if framer == None else framer
        framer._new_code_line(f'if {condition}:')
        true_block = true_block.replace("\n", "; ")
        framer._new_code_line(f'    {true_block}')
        if false_block:
            framer._new_code_line('else:')
            false_block = false_block.replace("\n", "; ")
            framer._new_code_line(f'    {false_block}')
    def to_last():
        s = System.framer
        System.framer = System.last_framer
        System.last_framer = s
class Var:
    def __init__(self, name, value, type: str = 'int', to_repr: bool = True, framer = None):
        framer = System.framer if framer == None else framer
        param_name = f'__tmp_{len(framer._code)}'
        self.name = param_name
        self.value = value
        with framer._lock:
            framer.var(param_name, value)
            framer._new_code_line(f'{name}: {type} = {repr(value) if to_repr else value}')
            framer._aliases[name] = param_name
def Get(name: str, framer: Framer | None = None):
    framer: Framer = System.framer if framer == None else framer
    return framer.get_thread_safe(name)
class Return:
    def __init__(self, value: Var, framer = None):
        framer = System.framer if framer == None else framer
        try: framer.var(f'__tmp_{len(framer._code) - 1}', f'{framer._vars.get(value.name)}')
        except AttributeError:
            raise FramerError(f'Exception in atribute parsing. \nObject [{value}, {type(value)}] has no atribute .name to create return. \nPlease, use [value] declaration like [`res = Var(...); Return(res, ...)`].')
class Code:
    def __init__(self, code: str, framer = None):
        framer = System.framer if framer == None else framer
        framer._new_code_line(code)
def Exec(framer = None):
    framer = System.framer if framer == None else framer
    with framer._lock:
        return framer.execute()
    
class Frame:
    def __init__(self, framer: str | Framer = 'new', safemode: bool = True, name: str = 'f1'):
        self.framer = Framer() if framer == 'new' else framer
        self.__safemode = safemode
        System.framers[name] = self.framer
    def Sys(self): 
        return System
    def Var(self, name, value, type: str = 'int', to_repr:bool = True):
        return Var(name, value, type, to_repr, self.framer)
    def Get(self, name: str): 
        return Get(name, self.framer)
    def Return(self, name: Var): 
        return Return(name, self.framer)
    def Code(self, code: str):
        return Code(code, self.framer)
    def Exec(self):
        if not self.__safemode: return Exec(self.framer)
        else: raise FrameError('Exec is not avialable in safemode.')
    def reset(self): self.framer = Framer()
    def __enter__(self): 
        self.framer.__enter__()
        return self
    def __exit__(self, *args, **kwargs): pass
    def __call__(self, *args, **kwds):
        return self.framer



Var('name', 'frame', framer=System.framers['basic'])

if __name__ == '__main__':
    with Frame() as f:
        x = Var('x', 10)
        y = Var('y', 50)
        System.match('x > y', 'print("x bigger")', 'print("y bigger")')
        res = Var('test', Get('x') * Get('y')) 
        Return(res)  
    print(Exec())  # → 500



'''
# Frame
The Frame - multitool module for programming.

Functions: 
    Framer functions -
        Framer, fExec, fGet, fVar, fSys, fReturn, fCode, @frmaing, framing_result
        FramerError, FramingError
'''

from .op import (Framer, Frame, Exec as fExec, Get as fGet, Var as fVar, 
                System as fSys, Return as fReturn, Code as fCode,
                FramerError)


def framing(
    framer: str | Framer = 'new',
    name_of_result_variable: str = 'res',
    return_frame: bool = False
    ):
    '''
## Frame decorator
### Args: 
arg {framer}:  object[Frame] | str = 'new' -

- Frame to load {name_of_result_variable}. 
- If == 'new', will be created new frame.
    
arg {name_of_result_variable}:  str = 'res' -
- Variable that will be created in {framer}.
    
arg {return_frame}:  bool = False -
- Args for choise create 'frame' variable (Frame object) in System.frames['temp'] .
### Example:
```
@frame_dec(return_frame=True)
def test():
    print('test')
    return 10
# geting frame object from decorator
res_frame = fGet('frame', fSys.framers['temp']) 
print(test(), res_frame)
# geting result variable from decorator
print(fGet('res', res_frame))
```
- Output:
```
10 <frame.op.Framer object at 0x10ddf3d10>
10
```
    '''
    if framer == 'new': framer_obj = Framer()
    else: framer_obj = framer
    def decorator(func):
        def wrapper(*args, **kwargs):
            res = func(*args, **kwargs)
            fVar(name_of_result_variable, res, framer=framer_obj)
            return res
        return wrapper
    if return_frame: fVar('frame', framer_obj, to_repr=False, framer=fSys.framers['temp'])
    return decorator



class FramingError(Exception):
    def __init__(self, *args):
        super().__init__(*args)
    
def framing_result(framer: Framer, func: object, name_of_result_variable: str = 'res', *func_args, **func_kwargs):
    '''
## Getting result from [@framing def ...] function
### Args:
- framer: object[Frame] -  framer to run.
- func: object - function for runing.
- name_of_result_variable: str - name_of_result_variable from decorator @framing.
- *args & **kwargs - arguments for [func] running.
### Example:
change 
```
# geting frame object from decorator
res_frame = fGet('frame', fSys.framers['temp'])
# runing 
print(test(), res_frame)
# geting result variable from decorator
print(fGet('res', res_frame))
```
to just
```
print(framing_result(fGet('frame', fSys.framers['temp']), test, 'res'))
```
    '''
    if isinstance(framer, Frame): framer = framer.framer
    resf = func(*func_args, **func_kwargs)
    resg = fGet(name_of_result_variable, framer)
    if resf != resg: raise FramingError(f'Variable [{name_of_result_variable}] is not found in Frame[{framer}].')
    return resg
